import numpy as np

from scorpionsql.functions import *
from scorpionsql.errfunc import *


class AggErr(object):
    
    def __init__(self, agg, keys, npts, errtype, metadata={}):
        """
        @param npts either an int or a tuple (min, max)
        """
        self.agg = agg
        self.keys = keys
        self.npts = npts
        self.errtype = ErrTypes(errtype, None)
        self.erreq = metadata.get('erreq', None)
        self.metadata = metadata or {}

    def __str__(self):
        return str([str(self.agg), self.keys, self.npts, str(self.errtype), self.erreq])

    error_func = property(lambda self: self.__error_func__())
    

    def __error_func__(self):
        f = self.agg.func
        f.set_errtype(self.errtype)
        return f
        if self.agg.func == 'avg':
            return FastAvgErrFunc(self)
        if self.agg.func == 'stddev':
            return FastStdErrFunc(self)
        if self.agg.func == 'sum':
            return FastSumErrFunc(self)
        if self.agg.func == 'min':
            return MinErrFunc(self)
        if self.agg.func == 'corr':
            return FastCorrErrFunc(self)
        raise NotImplementedError

    def bad_error_funcs(self, keys=None):
      if keys is None:
        keys = self.keys
      if self.errtype.errtype == ErrTypes.EQUALTO:
        ret = []
        for eqv in self.erreq:
          ef = self.agg.func.clone()
          ef.set_errtype(ErrTypes(self.errtype.errtype, eqv))
          ret.append(ef)
        return ret
      return [self.error_func.clone() for key in keys]

    def good_error_funcs(self, keys=[]):
      ret = []
      for key in keys:
        ef = self.error_func.clone()
        ef.errtype.errtype = ErrTypes.EQUALTO
        ef.errtype.erreq = None
        ret.append(ef)
      return ret






import psycopg2
from datetime import datetime, date

from scorpionsql.db import *

class Query(object):
  def __init__(self, db, select, fr, where,
      group=None, order=None, limit=None):
    self.db = db
    self.select = Select(select)
    self.fr = fr
    self.where = where
    self.group = group
    self.order = order
    self.limit = limit

  tables = property(lambda self: self.fr)

  def clone(self):
    select = Select(self.select)
    fr = list(self.fr)
    where = list(self.where)
    group = self.group and list(self.group) or None
    order = self.order and list(self.order) or None
    limit = self.limit and list(self.limit) or None
    return Query(self.db, select, fr, where, group, order, limit)


  def __str__(self):
    select = str(self.select)
    fr = ','.join(self.fr)
    where = self.wherestr
    
    sql = 'SELECT %s FROM %s' % (select, fr)
    if where:
        sql = '%s WHERE %s' % (sql, where)
    if self.group:
        sql = '%s GROUP BY %s' % (sql, ','.join(list(self.group)))
    if self.order:
        sql = '%s ORDER BY %s' % (sql, ','.join(list(self.order)))
    elif self.group:
        sql = '%s ORDER BY %s' % (sql, ','.join(list(self.group)))
    if self.limit is not None:
        sql = '%s LIMIT %s' % (sql, self.limit)
    return sql

  def __wherestr__(self, where=None):
    if where is None:
      where = self.where
    where = filter(bool, list(where))
    if not where: return None
    where = ["(%s)" % clause for clause in where]
    return " and ".join(where)
  wherestr = property(__wherestr__)


  def to_outer_join_sql(self, orig_where = []):
    """
    Construct a SQL query that includes all groups that would have
    existed (exists in the orig_where clause) if the WHERE clause were not present
    """
    if not self.group:
      return str(self)

    select = str(self.select)
    fr = ','.join(self.fr)
    filter_q = "SELECT * FROM %s" % fr
    if self.where:
      filter_q += " WHERE %s" % self.wherestr

    select_clause = self.select.coalesced_str()
    groupby = ','.join(list(self.group))
    if self.order:
      orderby = ','.join(list(self.order))
    else:
      orderby = groupby

    if orig_where:
      orig_where = self.__wherestr__(orig_where)
      orig_where = ' WHERE %s' % orig_where
    else:
      orig_where = ''

    sql = """with bar as (%s)
    select %s
    from 
      ((select distinct %s from %s %s) as foo
        left outer join  
      bar using (%s)) 
    GROUP BY %s
    ORDER BY %s;""" % (filter_q, select_clause, groupby, fr, orig_where, groupby, groupby, orderby)
    print sql
    return sql

    """
    with right as (select * from lqm where WHERECLAUSE)
    select coalesce(sum(total_cost), 0), ccm_payor 
    from 
      ((select distinct ccm_payor from lqm) as payors 
        left outer join  
      right using (ccm_payor)) 
    GROUP BY ccm_payor;
    """

  def prettify(self):
    select = str(self.select)
    fr = ','.join(self.fr)
    where = self.wherestr
    
    sql = 'SELECT %s\nFROM %s' % (select, fr)
    if where:
        sql = '%s\nWHERE %s' % (sql, where)
    if self.group:
        sql = '%s\nGROUP BY %s' % (sql, ','.join(list(self.group)))
    if self.order:
        sql = '%s\nORDER BY %s' % (sql, ','.join(list(self.order)))
    if self.limit is not None:
        sql = '%s\nLIMIT %s' % (sql, self.limit)
    return sql
    


  def get_filter_qobj(self, keys=None):
    """
    Return a copy of this Query object with additional where clauses for the 
    keys in the argument
    """
    # only care about columns in aggregates right?
    cols = set()
    for agg in self.select.aggregates:
        cols.update(agg.cols)
    sels = [SelectExpr(col, [col], col, None) for col in cols]
    select = Select(sels)
    where = list(self.where)
    if keys:
      keys = list(keys)
      keys = map(sqlize, list(keys))
      expr = self.select.nonaggs[0].expr

      clause = []
      if None in keys:
        clause.append("%s is null" % expr)
      if len([k for k in keys if k is not None]) > 0:
        clause.append("%s in %%s" % expr)
      clause = " or ".join(clause)
      where.append(clause)
    else:
      where.append( '%s = %%s' % (self.select.nonaggs[0].expr ) )
    q = Query(self.db, select, self.fr, where)
    return q
    

class Select(list):
  def __init__(self, *args):
    try:
      super(Select, self).__init__(*args)
    except Exception as e:
      print e
      print args
      super(Select, self).__init__(*[args])

  def __str__(self):
    return ', '.join(map(str, self))

  def coalesced_str(self):
    return ', '.join([s.coalesced_str() for s in self])
      
  aggregates = property( lambda self: filter(lambda e: isinstance(e, SelectAgg), self) )
  aggs = property( lambda self: filter(lambda e: isinstance(e, SelectAgg), self) )    
  nonaggs = property( lambda self: filter(lambda e: not isinstance(e, SelectAgg), self) )


class SelectAgg(object):
  def __init__(self, alias, func, cols, expr, attr):
    self.alias = str(alias) if alias else None # alias, if any
    self._func = func # aggregate function
    self.cols = map(str, cols) # list of attributes 
    self.expr = expr # string of expression passed into aggregate function
    self.attr = attr # actual attribute token from sqlparse
    self.fname = attr[0]
    self.isagg = True

    #if not self.alias:
    #self.alias = str(self.fname)

  shortname = property(lambda self: self.alias or self.fname or self.expr)
  func = property(lambda self: self._func.clone())

  def __str__(self):
    s = self.expr
    if self.alias:
      s = '%s as %s' % (s, self.alias)
    return s

  def coalesced_str(self):
    s = "coalesce(%s, 0)" % self.expr
    if self.alias:
      s = "%s as %s" % (s, self.alias)
    else:
      s = "%s as %s" % (s, self.fname)
    return s

    
class SelectExpr(object):
  def __init__(self, alias, cols, expr, attr):
    self.alias = str(alias) if alias else None
    self.cols = map(str, cols)
    self.expr = expr
    self.attr = attr
    self.isagg = False

  shortname = property(lambda self: self.alias or self.expr)

  def __str__(self):
    if self.alias:
        return '%s as %s' % (self.expr, self.alias)
    return self.expr

  def coalesced_str(self):
    return str(self)


def quote_sql_str(s):
  return psycopg2.extensions.QuotedString(s).getquoted()

def sqlize(val):
  if val is None:
    return val
  if isinstance(val, basestring):
    return quote_sql_str(val)
  if isinstance(val, datetime):
    return quote_sql_str(str(val))
    if val.hour == 0 and val.minute == 0 and val.second == 0:
      return '%s::date' % quote_sql_str(str(val))
    return '%s::timestamp' % quote_sql_str(str(val))
  if isinstance(val, date):
    return quote_sql_str(str(val))
    return '%s::date' % quote_sql_str(str(val))
  return str(val)




if __name__ == '__main__':

    '(((avg(temp))) / (std(temp) + 2))'

    import pyparsing
    from pyparsing import *

        

    exprStack = []

    def pushNum(s, loc, toks):
        print 'num ', toks
    def pushVar(s, loc, toks):
        print 'var ', toks
    def pushNest(s, loc, toks):
        print 'nest ', toks
    def pushFunc(s, loc, toks):
        print 'func ', toks
    def pushArith(s, loc, toks):
        print 'arith ', toks

    def pushFirst( strg, loc, toks ):
        exprStack.append( toks[0] )
    def pushUMinus( strg, loc, toks ):
        if toks and toks[0]=='-': 
            exprStack.append( 'unary -' )
            #~ exprStack.append( '-1' )
            #~ exprStack.append( '*' )


    Number = Group(Optional("-") + Word(nums) + Optional( "." + Optional(Word(nums)))).setParseAction(pushNum)
    Var = Word(alphas).setParseAction(pushVar)
    expr = Forward()
    #foo = Forward()
    nest = Group("(" + expr + ")").setParseAction(pushNest)
    func = Group(Word(alphas) + "(" + expr + ")").setParseAction(pushFunc)
    op = Word("+_/*", max=1)

    stuff = ( func | nest | Number | Var )
    expr << (  (stuff + op + expr).setParseAction(pushArith) | stuff | Empty() )
    #foo << ( (Word("+-/*", max=1) + expr + foo) | Empty()  )
    
    res = expr.parseString("1 + 2 / foo(bar+2)")

    import pdb
    pdb.set_trace()



import numpy as np
from scorpionsql.functions import *


def compute_bad_inf(bdelta, bcount, c):
  if bcount == 0:
    return 0
  return float(bdelta / pow(bcount, c))
compute_bad_inf = np.vectorize(compute_bad_inf)

def compute_bad_score(bds, bcs, c, smooth=0.025):
  bcs = np.array(bcs).astype(float)

  if not bds or not bcs.any():
    return c * -float('inf')

  topbots = zip(bds, bcs)
  binfs = [compute_bad_inf(top, bot, c) for top, bot in topbots]
  binfs = np.array(binfs)

  bcs[bcs > 0] = 1
  bcs[bcs <= 0] = smooth
  weights = bcs / bcs.sum()
  try:
    return np.average(binfs, axis=0, weights=weights)
  except:
    print bcs
    print weights
    print bds
    import traceback
    traceback.print_exc()
    import pdb
    pdb.set_trace()

def compute_influence(l, binf, ginf, nclauses=0, *args):
  return l * binf - (1. - l) * ginf - 0.0001 * nclauses # XXX: add - len(rule.clauses)


class ErrTypes(object):
    WRONG = 0
    EQUALTO = 1
    TOOHIGH = 2
    TOOLOW = 3

    def __init__(self, errtype, erreq=None):
        try:
            self.errtype = errtype.errtype
        except:
            self.errtype = errtype
        self.erreq = erreq            
        #if self.errtype == ErrTypes.EQUALTO and erreq is None:
        #    raise "Error Type EQUALTO needs a value"

    def __str__(self):
        return '%d\t%s' % (self.errtype, self.erreq)

    @staticmethod
    def valid(err):
        return err >= 0 and err <= 3

    def __call__(self, oldv, newv):
      if self.errtype == ErrTypes.TOOHIGH:
          return oldv - newv
      if self.errtype == ErrTypes.TOOLOW:
          return newv - oldv
      if self.errtype == ErrTypes.EQUALTO:
        diff = abs(self.erreq - newv)
        olddiff = abs(self.erreq - oldv)
        return 1. - (1.+diff)/(1.+olddiff)
      return abs(newv - oldv)

    def clone(self):
        return ErrTypes(self.errtype, erreq=self.erreq)

def setup_wrapper(f):
    def h(self, table):
        ret = f(self, table)
        if self.errtype.errtype == ErrTypes.EQUALTO:
          if self.errtype.erreq is None:
            self.errtype.erreq = ret
        return ret
    return h

class BaseErrFunc(object):
    """
    Function is
    initialized with child nodes

    __call__(data) passes data to child nodes
    """
    def __init__(self):
        self.ncalls = 0
        self.errtype = None
        self.data = None
        self.parent = None
        self.children = []

    is_root = property(lambda self: self.parent == None)

    def clone(self):
        pass

    def set_errtype(self, errtype):
        if errtype:
            self.errtype = errtype.clone()

    def setup(self, table):
        self.domain = table.domain
        self.data = table

    def state(self, data):
        raise RuntimeError("not implemented")

    def update(self, states, num=1):
        raise RuntimeError("not implemented")

    def remove(self, big_state, state, n=1):
        raise RuntimeError("not implemented")

    def recover(self, state):
        raise RuntimeError("not implemented")

    def __call__(self, data, opt):
        pass

    def distribution(self, table, k=0.05):
        k = max(len(table) * k, min(len(table), 10))
        sampler = Orange.data.sample.SubsetIndices2(p0=len(table) - k)
        sampler.random_generator = Orange.misc.Random(len(table))

        scores = []
        for i in xrange(50):
            idxs = sampler(table)
            sample = table.select_ref(idxs).to_numpyMA('ac')[0]
            score = self(sample, "rm") / k
            scores.append( score )

        return np.mean(scores), np.std(scores)    
    

class AggErrFunc(BaseErrFunc):
    def __init__(self, children, klass):
        BaseErrFunc.__init__(self)

        self.klass = klass
        self.f = klass()
        self.fempty = klass()
        self.children = children
        for c in children:
            c.parent = self

    def clone(self):
        newchildren = [c.clone() for c in self.children]
        ret = type(self)(newchildren)
        ret.set_errtype(self.errtype)
        return ret

    def set_errtype(self, errtype):
        BaseErrFunc.set_errtype(self, errtype)
        map(lambda c: c.set_errtype(errtype), self.children)
        return self

    @setup_wrapper
    def setup(self, table):
        BaseErrFunc.setup(self, table)
        domain = table.domain
        vals = map(lambda c: c.setup(table), self.children)
        self.f(*vals)
        self.value = self.f.value()
        return self.value

    def state(self, data):
        vals = map(lambda c: c(data, 'scratch'), self.children)
        return self.f.state(*vals)

    def update(self, states, num=1):
        return self.f.update(states, num=num)

    def remove(self, big_state, state, n=1):
        return self.f.remove(big_state, state, n)

    def recover(self, state):
        "recompute the actual value, then compare it against the truth"
        newval = self.f.recover(state)
        return self.errtype(self.value, newval)


    def __call__(self, data, mode='rm'):
        self.ncalls += 1
        vals = map(lambda c: c(data, mode), self.children)
        if mode == 'rm':
            newval = self.f.delta(rm=vals)
        if mode == 'scratch':
            newval = self.fempty.delta(add=vals)

        if self.is_root:
            return self.errtype(self.value, newval)
        else:
            return newval


class CorrErrFunc(AggErrFunc):
    def __init__(self, children):
        AggErrFunc.__init__(self, children, LinearFit)

    @setup_wrapper
    def setup(self, table):
        BaseErrFunc.setup(self, table)
        domain = table.domain
        vals = map(lambda c: c.setup(table), self.children)
        self.f(*vals)
        self.value = self.f.corr()
        return self.value

    def __call__(self, data, mode='rm'):
        self.ncalls += 1
        vals = map(lambda c: c(data, mode), self.children)
        if mode == 'rm':
            newval = self.f.corr(rm=vals)
        if mode == 'scratch':
            newval = self.fempty.corr(add=vals)

        if self.is_root:
            return self.errtype(self.value, newval)
        return newval

        

class StdErrFunc(AggErrFunc):
    def __init__(self, children):
        AggErrFunc.__init__(self, children, StdFunc)

class AvgErrFunc(AggErrFunc):
    def __init__(self, children):
        AggErrFunc.__init__(self, children, AvgFunc)

class MinErrFunc(AggErrFunc):
    def __init__(self, children):
        AggErrFunc.__init__(self, children, MinFunc)

class MaxErrFunc(AggErrFunc):
    def __init__(self, children):
        AggErrFunc.__init__(self, children, MaxFunc)

class SumErrFunc(AggErrFunc):
    def __init__(self, children):
        AggErrFunc.__init__(self, children, SumFunc)

class CountErrFunc(AggErrFunc):
    def __init__(self, children):
        AggErrFunc.__init__(self, children, CountFunc)


def _add(a,b):
    return a + b
def _sub(a,b):
    return a - b
def _mul(a,b):
    return a / b
def _div(a,b):
    return a / b


class ArithErrFunc(BaseErrFunc):
    op2f = {'+' : _add,
            '-' : _sub,
            '*' : _mul,
            '/' : _div}

    def __init__(self, op, left, right):
        BaseErrFunc.__init__(self)
        self.op = op
        self.l = left
        self.r = right
        self.f = ArithErrFunc.op2f[op]
        self.l.parent = self
        self.r.parent = self

    def clone(self):
        ret = ArithErrFunc(self.op, self.l.clone(), self.r.clone())
        ret.set_errtype(self.errtype)
        return ret

    def set_errtype(self, errtype):
        BaseErrFunc.set_errtype(self, errtype)
        map(lambda c: c.set_errtype(errtype), [self.l, self.r])

    @setup_wrapper
    def setup(self, table):
        BaseErrFunc.setup(self, table)
        domain = table.domain
        self.lval = self.l.setup(table)
        self.rval = self.r.setup(table)
        self.value = self.f(self.lval, self.rval)
        return self.value

    def __call__(self, data, mode='rm'):
        self.ncalls += 1
        vals = map(lambda c: c(data, mode), [self.l, self.r])
        newval = self.f(*vals)

        if self.is_root:
            return self.errtype(self.value, newval)
        return newval

class AbsErrFunc(BaseErrFunc):
    def __init__(self, children):
        BaseErrFunc.__init__(self)
        self.child = children[0] if isinstance(children, list) else children
        self.child.parent = self

    def clone(self):
        ret = AbsErrFunc(self.child.clone())
        ret.set_errtype(self.errtype)
        return ret

    @setup_wrapper
    def setup(self, table):
        v = abs(self.child.setup(table))
        self.value = v
        return v

    def __call__(self, data, mode='rm'):
        self.ncalls += 1
        newval = abs(self.child(data, mode))
        
        if self.is_root:
            return self.errtype(self.value, newval)
        return newval

    


class Constant(BaseErrFunc):
    def __init__(self, v):
        BaseErrFunc.__init__(self)        
        self.v = v

    def clone(self):
        return self

    @setup_wrapper
    def setup(self, table):
        return self.v
        return np.array([self.v] * len(table))

    def __call__(self, data_array, mode='rm'):
        self.ncalls += 1
        return self.v
        return np.array([self.v] * data_array.shape[0])

class Var(BaseErrFunc):
    def __init__(self, var):
        BaseErrFunc.__init__(self)
        self.var = var
        self.idx = None

    def clone(self):
        return self

    def setup(self, table):
        domain = table.domain
        self.idx = domain.index(domain[self.var])
        return self(table.to_numpyMA('ac')[0])
    
    def __call__(self, data_array, mode='rm'):
        self.ncalls += 1
        if isinstance(data_array, np.ndarray):
            if len(data_array.shape) == 1:
                return data_array[self.idx]
            return data_array[:, self.idx]
        return np.array([float(row[self.idx].value) for row in data_array])

class DummyVar(BaseErrFunc):
    def __call__(self):
        BaseErrFunc.__init__(self)

    def clone(self):
        return self

    def setup(self, data):
        return np.array(sum(data) )

    def __call__(self, data_array, mode='rm'):
        return np.array(data_array)



if __name__ == '__main__':
    ef = AvgErrFunc([DummyVar()])
    s1 = ef.state(range(10))
    s2 = ef.state(range(4))
    print ef.recover(s1), np.mean(range(10))
    print ef.recover(s2), np.mean(range(4))
    print ef.recover(ef.update(s1, s2)), np.mean(range(10) + range(4))
    print ef.recover(ef.remove(s1, s2)), np.mean(range(4,10))



import math
import numpy as np
from itertools import repeat

NAN = float('NaN')

class Func(object):
    def __init__(self, data=None):
        pass

    def state(self, data):
        pass

    def update(self, states, num=1):
        pass

    def remove(self, big_state, state, num=1):
        pass

    def recover(self, state):
        pass

    def __call__(self, data):
        pass

    def value(self):
        pass
    
    def delta(self, add=None, rm=None, update=False):
        pass

class CompositeFunc(object):
    def __init__(self, children, f, data=None):
        """
        @param children the aggregate functions that are composed together
        @param f output of children -> single value
        """
        self.children = children
        self.f = f

        if data:
            self(data)

    def value(self):
        return self.f(*map(lambda c: c.value(), self.children))

    def __call__(self, data):
        if data:
            map(lambda c: c(data), self.children)
        return self.value()

    def delta(self, **kwargs):
        return self.f(*map(lambda c: c.delta(**kwargs), self.children))

class StdFunc(Func):
    def __init__(self, data=None):
        self.x2 = 0.
        self.n = 0.
        self.mean = 0.
        self.std = 0.
        if data is not None:
            self(data)

    def state(self, data):
        if len(data.shape) == 2:
            data = data.reshape(data.shape[0])
        x2 = np.dot(data, data)
        n = len(data)
        total = data.sum()

        return (x2, total, n)


    def update(self, states, num=1):
        x2s, totals, counts = zip(*states)
        x2 = sum(x2s)
        count = sum(counts)
        total = sum(totals)
        return (x2*num, total*num, count*num)
    
    def remove(self, bstate, state, num=1):
        x2, total, n = tuple(bstate)
        x2p, tp, np = self.update((state,))
        return (x2-(num*x2p), total-(num*tp), n-(num*np))
   
    def recover(self, state):
        x,t,n = tuple(state)
        top = (n*x - t**2)
        bot = n*n
        if top < -1e-9 or bot == 0:
            return NAN
        return math.sqrt(top/bot)


    def value(self):
        return self.std
        
    def __call__(self, data):
        if len(data.shape) == 2:
            data = data.reshape(data.shape[0])
        self.x2 = np.dot(data, data)
        self.n = len(data)
        self.mean = np.mean(data)
        self.std = np.std(data)

    def delta(self, add=None, rm=None, update=False):
        add = add[0] if add else add
        rm = rm[0] if rm else rm
        #add = add.reshape(add.shape[0]) if add is not None and len(add.shape)==2 else add
        #rm = rm.reshape(rm.shape[0]) if rm is not None and len(rm.shape)==2 else rm

        x2, n, mean, std = self.x2, self.n, self.mean, self.std
        total = mean * n
        if add is not None:
            x2 += np.dot(add, add)
            total += add.sum()
            n += len(add)            
        if rm is not None:
            x2 -= np.dot(rm, rm)
            total -= rm.sum()
            n -= len(rm)
        top = (n*x2 - total**2)
        bot = n*n
        if top < -1e-9:
            top = 0.
        if bot == 0 or top < 0:
            return NAN
        std = math.sqrt( top / bot )

        if update:
            self.x2 = x2
            self.n = n
            self.mean = total / n
            self.std = std
        
        return std

class AvgFunc(Func):
    def __init__(self, data=None):
        self.n = 0.
        self.total = 0.
        if data is not None:
            self(data)

    def value(self):
        return self.total / self.n

    def state(self, data):
        if len(data.shape) == 2:
            data = data.reshape(data.shape[0])
        return (np.mean(data), len(data))

    def update(self, states, num=1):
        totals, counts = zip(*states)
        totaln = float(sum(counts))
        newavg = sum([t * count / totaln for t,count in zip(totals, counts)])
        return (newavg, num*totaln)

    def remove(self, bstate, state, num=1.):
        t, count = tuple(bstate)
        rmt, rmcount = self.update((state,), num=num)
        newcount = float(count - rmcount)
        if newcount == 0:
            return (0, 0)
        newt = t*count / newcount - rmt*rmcount / newcount 
        return (newt, newcount)

    def recover(self, state):
        t,n = tuple(state)
        if not n:
            return NAN
        return t 

    def __call__(self, data):
        if len(data.shape) == 2:
            data = data.reshape(data.shape[0])
        self.total = data.sum()
        self.n = len(data)

    def delta(self, add=None, rm=None, update=False):
        add = add[0] if add else add
        rm = rm[0] if rm else rm
        #add = add.reshape(add.shape[0]) if add is not None and len(add.shape)==2 else add
        #rm = rm.reshape(rm.shape[0]) if rm is not None and len(rm.shape)==2 else rm

        total, n = self.total, self.n
        if add is not None:
            total += add.sum()
            n += len(add)
        if rm is not None:
            total -= rm.sum()
            n -= len(rm)
        if n == 0:
            return 0
            #return NAN

        if update:
            self.total = total
            self.n = n
        
        return total / float(n)


class MinFunc(Func):
    def __init__(self, data=None):
        self.min = None
        if data is not None:
            self(data)        

    def value(self):
        return self.min

    def __call__(self, data):
        if len(data.shape) == 2:
            data = data.reshape(data.shape[0])
        self.min = min(data)

    def delta(self, add=None, rm=None, update=False):
        ret = self.min
        if add is not None:
            ret = min(ret, min(add))
        if rm is not None:
            raise
        if update:
            self.min = ret
        return ret

class MaxFunc(Func):
    def __init__(self, data=None):
        self.max = None
        if data is not None:
            self(data)        

    def value(self):
        return self.max

    def __call__(self, data):
        if len(data.shape) == 2:
            data = data.reshape(data.shape[0])
        self.max = max(data)

    def delta(self, add=None, rm=None, update=False):
        ret = self.max
        if add is not None:
            ret = max(ret, max(add))
        if rm is not None:
            raise

        if update:
            self.max = ret
        return ret


class SumFunc(Func):
    def __init__(self, data=None):
        self.total = 0.
        if data is not None:
            self(data)        

    def value(self):
        return self.total



    def state(self, data):
        if len(data.shape) == 2:
            data = data.reshape(data.shape[0])
        return (data.sum(), len(data))

    def update(self, states, num=1):
        sums, counts = zip(*states)
        return (sum(sums) * num, sum(counts) * num)

    def remove(self, bstate, state, num=1):
        return (bstate[0] - num*state[0], bstate[1]-state[1])

    def recover(self, state):
        return state[0]

    def __call__(self, data):
        if len(data.shape) == 2:
            data = data.reshape(data.shape[0])
        self.total += data.sum()

    def delta(self, add=None, rm=None, update=False):
        ret = self.total
        if add is not None:
            ret += sum(el.sum() for el in add)
        if rm is not None:
            ret -= sum(el.sum() for el in rm)

        if update:
            self.total = ret
        return ret



class CountFunc(Func):
    def __init__(self, data=None):
        self.total = 0.
        if data is not None:
            self(data)        

    def value(self):
        return self.total

    def state(self, data):
        if len(data.shape) == 2:
            data = data.reshape(data.shape[0])
        return (len(data), )

    def update(self, states, num=1.):
        return (sum(s[0] for s in states) * num)

    def remove(self, bstate, state, num=1):
        return (bstate[0] - num*state[0],)

    def recover(self, state):
        return state[0]


    def __call__(self, data):
        if len(data.shape) == 2:
            data = data.reshape(data.shape[0])
        self.total = len(data)

    def delta(self, add=None, rm=None, update=False):
        ret = self.total
        if add is not None:
            ret += sum(map(len,add))
        if rm is not None:
            ret -= sum(map(len, rm))

        if update:
            self.total = ret
        return ret




if __name__ == '__main__':
    
    stdf = StdFunc()
    data = range(200)
    stdf(np.array(data))
    print stdf.delta(rm=np.array(data[100:])), np.std(data[:100])



import subprocess
import os
import json
import sys

from scorpionsql.errfunc import *
from scorpionsql.functions import *
from scorpionsql.sql import *

dirname = os.path.dirname(os.path.abspath( __file__ ))
classes = "%s/jars/scala-library.jar:%s/jars/sqlparser.jar" % (dirname, dirname)
__agg2f__ = {'avg' : AvgErrFunc,
             'std' : StdErrFunc,
             'stddev' : StdErrFunc, 
             'stddev_samp': StdErrFunc,
             'stddev_pop': StdErrFunc,
             'min' : MinErrFunc,
             'max' : MaxErrFunc,
             'sum' : SumErrFunc,
             'corr' : CorrErrFunc,
             'count' : CountErrFunc,
             'abs' : AbsErrFunc}

def parse_sql(q):
    return SqlParser()(q)

class SqlParser(object):
    def __init__(self):
        pass

    def __call__(self, q):
        jsonobj = self.fetch_json(q)
        if not jsonobj:
            raise "could not parse"

        queryobj = self.json_to_obj(jsonobj)
        return queryobj

    def fetch_json(self, q):
        try:
            cmd = ["java", "-cp", classes, "main", q]
            ret = subprocess.check_output(cmd)
            print ret
            return json.loads(ret)
        except subprocess.CalledProcessError as e:
            print >>sys.stderr, str(e)
            import traceback
            traceback.print_exc()
            return None


    def json_to_obj(self, j):
        select = j['select']
        fr = j['from']
        where = j['where']
        group = j['group']
        having = j['having']
        order = j['order']
        limit = j['limit']


        aggsels, normsels = [], []
        sels = []
        import pdb
        #pdb.set_trace()
        for sexpr, alias in select:
            cols = get_vars(sexpr)
            expr_str = expr_to_str(sexpr)

            if alias is None:
                if sexpr[0] == '_var_':
                    alias = sexpr[1]
                elif sexpr[0].strip('_') not in ('literal', 'cast', 'extract', '+', '-', '*', '/'):
                    alias = sexpr[0]
            
            if alias in group:
                sel = SelectExpr(alias, cols, expr_str, sexpr)
            else:
                func = construct_func_expr(sexpr)
                sel = SelectAgg(None, func, cols, expr_str, sexpr)
            sels.append(sel)

        order = ['%s %s' % (a,b) for a,b in map(tuple, order)]
        limit = limit[0] if limit else None

        return Query(None,
                     sels,
                     fr,
                     where,
                     group,
                     order,
                     limit)


def expr_to_str(n, l=None):
    """
    construct SQL string from expression node
    """
    op = n[0]
    if op.startswith('_') and op.endswith('_'):
        op = op.strip('_')
        if op == 'var':
            return n[1]
        elif op == 'literal':
            if isinstance(n[1], basestring):
                return "'%s'" % n[1]
            return str(n[1])
        elif op == 'cast':
            return "(%s)::%s" % (expr_to_str(n[1]), n[2])
    elif op in '+-*/':
        return "(%s) %s (%s)" % (expr_to_str(n[1]), op, expr_to_str(n[2]))
    elif op == "extract":
        return "extract( %s from %s )" % (n[1], expr_to_str(n[2]))
    else:
        arg = ','.join(map(expr_to_str, n[1:]))
        return "%s(%s)" % (op, arg)
    
def get_vars(n):
    """
    extract variables from expression node
    defined by tuple-pair: (_var_, [variable name])
    """
    op = n[0]
    if op.startswith('_') and op.endswith('_'):
        op = op.strip('_')
        if op == 'var':
            return [n[1]]
        return []
    else:
        ret = []
        for c in n[1:]:
            vs = get_vars(c)
            if vs:
                ret.extend(vs)
        return ret


def construct_func_expr(n):
    """
    construct the function expression 
    """
    op = n[0]
    if op.startswith('_') and op.endswith('_'):
        op = op.strip('_')
        if op == 'var':
            return Var(str(n[1]))
        elif op == 'literal':
            if isinstance(n[1], basestring):
                raise "not implemented"
            return Constant(n[1])
        elif op == 'cast':
            raise "not implemented"
    elif op in '+-/*':
        return ArithErrFunc(op, *map(construct_func_expr,n[1:]))
    else:
        klass = __agg2f__.get(op, None)
        if klass:
            return klass(map(construct_func_expr, n[1:]))
        raise "no klass"


if __name__ == '__main__':
    from tests.gentestdata import *
    for name in datasetnames:
        test_data = get_test_data(name)
        q = test_data[1]
        parser = SqlParser()
        qobj = parser(q)
        print qobj
    exit()
    import Orange
    t = Orange.data.Table('iris')
    names = ['slen', 'swid', 'plen', 'pwid', 'iris']
    for d, n in zip(t.domain, names):
        d.name = n

    func = qobj.select.aggregates[0].func
    func.setup(t, ErrTypes(ErrTypes.TOOHIGH))

    tn = t.to_numpy('a')[0]
    for i in xrange(len(t)):
        print func(tn[(i,),:], "rm"), tn[i,:]
        
    



#!/usr/bin/env python
try:
    from setuptools import setup, find_packages
except ImportError:
    import ez_setup
    ez_setup.use_setuptools()
from setuptools import setup, find_packages
import scorpionsql

setup(name="scorpionsql",
    version=scorpionsql.__version__,
    description="sql packaged used by scorpion and summary",
    license="MIT",
    author="Eugene Wu",
    author_email="eugenewu@mit.edu",
    url="http://github.com/sirrice/scorpionsql",
    include_package_data = True,      
    packages = find_packages(),
    package_dir = {'scorpionsql' : 'scorpionsql'},
    scripts = [ ],
    package_data = {
      'scorpion': [
        'jars/scala-library.jar',
        'jars/sqlparser.jar'
      ]
    },
    install_requires = [
      'psycopg2', 'sqlalchemy', 'numpy', 'pyparsing'
    ],
    keywords= "")
